;;; auto-correct.el --- Support for auto-correction -*- lexical-binding: t; -*-

;; Copyright (C) 2017 Ian Dunn

;; Author: Ian Dunn <dunni@gnu.org>
;; Maintainer: Ian Dunn <dunni@gnu.org>
;; Keywords: editing
;; Version: 1.0
;; Created: 11 May 2016
;; Modified: 11 May 2016

;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; Runs auto-correct for Emacs.  By default, nothing is corrected.  One must
;; "train" auto-correct before any words are fixed.

;; To train it, use M-x `auto-correct-fix-and-add' with point on a misspelled
;; word.  This will create a "fix" for the typo.  From then on, whenever you
;; misspell that word the same way, `auto-correct-mode' will fix it for you.

;; Expansion is case-insensitive, so trying to fix alice as Alice won't work.

;; Behind the scenes, auto-correct uses an abbrev table, so in order to clean
;; out or modify any fixes auto-correct has learned, use `list-abbrevs'.  This
;; also means that fixes are saved between Emacs sessions along with the abbrev
;; tables.

;; To add fixes for all misspelled words in a buffer, use
;; `auto-correct-scan-buffer'.

;;; Code:

(eval-when-compile (require 'subr-x))

(defgroup auto-correct nil
  "Auto correction support."
  :prefix "auto-correct-"
  :group 'editing)

(defvar-local auto-correct-predicate nil
  "Predicate to check for auto-correct abbrevs.")

(defun auto-correct-expand-p ()
  "Return non-nil if auto-correct should operate on the current point.

To customize this behavior, set `auto-correct-predicate'."
  (when auto-correct-predicate (funcall auto-correct-predicate)))

(define-abbrev-table 'auto-correct-abbrev-table nil
  "Abbrev table where automatic corrections are stored."
  :enable-function #'auto-correct-expand-p)

;;;###autoload
(defun auto-correct-fix-and-add (p)
  "Use `ispell-word' to fix a misspelled word at point.

Once the misspelled word is fixed, auto-correct will remember the
fix and auto-correct it from then on, so long as
`auto-correct-mode' is enabled.

With a non-nil argument P (interactively, the prefix argument),
create a fix for the typo that will be auto-corrected for buffers
using the current local mode."
  (interactive "P")
  (when-let ((word-before (thing-at-point 'word))
             (correction (ispell-word nil 'quiet)))
    (when (and correction (consp correction))
      ;; The correction was entered by hand.
      (setq correction (car correction)))
    (if (and (not (or (eq correction 0)  ;; Word was corrected from list
                    (eq correction 'quit))) ;; Session was exited
             (not (equal word-before correction))) ;; Word was corrected
        (let ((bef (downcase word-before))
              (aft (downcase correction)))
          (define-abbrev
            (if p local-abbrev-table auto-correct-abbrev-table)
            bef aft nil :count 1)
          ;; Save the abbrevs.
          (write-abbrev-file)
          (message "\"%s\" now expands to \"%s\"" bef aft))
      (user-error "No correction made at or before point"))))

;;;###autoload
(defun auto-correct-scan-buffer ()
  "Scan current buffer for misspelled words.

When a misspelled word is found, offer to correct the misspelled
word and auto-correct the typo in the future."
  (interactive)
  (save-excursion
    (goto-char (point-min))
    (while (forward-word)
      (auto-correct-fix-and-add nil))
    (ispell-pdict-save)))

(defun auto-correct-scan-region (start end)
  "Scan the region between START and END for misspelled words.

Interactively, START and END are the current region.

When a misspelled word is found, offer to correct the misspelled
word and auto-correct the typo in the future."
  (interactive "r")
  (save-restriction
    (narrow-to-region start end)
    (auto-correct-scan-buffer)))

;;;###autoload
(defun auto-correct-scan ()
  "Scan the buffer or region for misspelled words.

When a misspelled word is found, offer to correct the misspelled
word and auto-correct the typo in the future."
  (interactive)
  (if (region-active-p)
      (auto-correct-scan-region (region-beginning) (region-end))
    (auto-correct-scan-buffer)))

;;;###autoload
(define-minor-mode auto-correct-mode
  "Minor mode for auto-correct.

Auto correct expansions will only work when this mode is enabled,
but auto-correct can be trained with `auto-correct-fix-and-add'
even if this mode is disabled.

\\{auto-correct-map}"
  :group 'auto-correct
  :global t
  :init-value nil
  :lighter " Auto-Correct"
  :keymap
  '(([remap (ispell-word)] . auto-correct-fix-and-add)
    ([remap (ispell)]      . auto-correct-scan)))

(add-to-list 'abbrev-minor-mode-table-alist
             `(auto-correct-mode ,auto-correct-abbrev-table)
             'append
             #'equal)

(provide 'auto-correct)

;;; auto-correct.el ends here
